from . import sync  # noqa: F401: apply magic async/sync conversion
from .coins import BCH, BSTY, BTC, COINS, GZRO, LTC, XRG  # noqa: F401
from .errors import errors
from .manager import APIManager

__all__ = list(COINS.keys()) + ["COINS", "APIManager", "errors"]



import asyncio
from collections import UserDict, defaultdict
from functools import partial
from typing import TYPE_CHECKING, Any, Callable, Dict, Iterable, Optional
from urllib.parse import urljoin

from bitcart.errors import NoCurrenciesRegisteredError

from .coins import COINS
from .event_delivery import EventDelivery

if TYPE_CHECKING:
    from aiohttp import ClientWebSocketResponse

    from .coin import Coin
    from .providers.jsonrpcrequests import RPCProxy


class CustomDict:
    def __getattr__(self, name: str) -> Any:
        return self.__getitem__(name)


class ExtendedDefaultDict(defaultdict, CustomDict):
    pass


class ExtendedDict(UserDict, CustomDict):
    pass


class APIManager(EventDelivery):
    def __init__(self, wallets: Dict[str, Iterable[str]] = {}):
        super().__init__()
        self.wallets = ExtendedDefaultDict(
            lambda: ExtendedDict(),
            {currency: self.load_wallets(currency, wallets) for currency, wallets in wallets.items()},
        )
        self.event_handlers = {}

    @classmethod
    def load_wallets(cls, currency: str, wallets: Iterable[str]) -> ExtendedDict:
        return ExtendedDict({wallet: cls.load_wallet(currency, wallet) for wallet in wallets})

    @classmethod
    def load_wallet(cls, currency: str, wallet: Optional[str]) -> "Coin":
        return COINS[currency](xpub=wallet)

    def add_wallet(self, currency: str, wallet: str) -> None:
        self.wallets[currency][wallet] = self.load_wallet(currency, wallet)

    def add_wallets(self, currency: str, wallets: Iterable[str]) -> None:
        self.wallets[currency].update(self.load_wallets(currency, wallets))

    def __getitem__(self, key: str) -> Any:
        return self.wallets.__getitem__(key)

    def __getattr__(self, key: str) -> Any:
        return self.__getitem__(key)

    def _merge_event_handlers(self, wallet: "Coin") -> None:
        wallet.event_handlers.update(self.event_handlers)

    async def _register_wallets(self, ws: "ClientWebSocketResponse") -> None:
        pass  # listen on all wallets

    def _get_websocket_server(self, currency: str) -> "RPCProxy":
        try:
            coin = next(iter(self.wallets[currency].values()))
        except StopIteration:  # pragma: no cover
            coin = self.load_wallet(currency, None)
        return coin.server  # type: ignore

    async def _start_websocket_for_currency(self, currency: str, reconnect_callback: Optional[Callable] = None) -> None:
        if reconnect_callback:
            reconnect_callback = partial(reconnect_callback, currency)
        server = self._get_websocket_server(currency)
        async with server.session.ws_connect(urljoin(server.url, "/ws")) as ws:
            await self._start_websocket_processing(ws, reconnect_callback=reconnect_callback)

    async def start_websocket_for_currency(
        self,
        currency: str,
        reconnect_callback: Optional[Callable] = None,
        force_connect: bool = False,
        auto_reconnect: bool = True,
    ) -> None:
        await self._websocket_base_loop(
            partial(self._start_websocket_for_currency, currency=currency),
            reconnect_callback=reconnect_callback,
            force_connect=force_connect,
            auto_reconnect=auto_reconnect,
        )

    async def start_websocket(
        self, reconnect_callback: Optional[Callable] = None, force_connect: bool = False, auto_reconnect: bool = True
    ) -> None:
        tasks = []
        for currency in self.wallets:
            tasks.append(
                self.start_websocket_for_currency(
                    currency, reconnect_callback=reconnect_callback, force_connect=force_connect, auto_reconnect=auto_reconnect
                )
            )
        await asyncio.gather(*tasks)
        if not tasks:
            raise NoCurrenciesRegisteredError(NoCurrenciesRegisteredError.__doc__)

    async def process_updates(
        self, updates: Iterable[dict], currency: Optional[str] = None, wallet: Optional[str] = None
    ) -> None:
        wallet_obj = self.wallets[currency].get(wallet)
        if not wallet_obj:
            wallet_obj = self.load_wallet(currency, wallet)  # type: ignore
        self._merge_event_handlers(wallet_obj)
        await wallet_obj.process_updates(updates, pass_instance=True)



import subprocess


def run_shell(args=[], timeout=30):
    process = subprocess.Popen(["tests/regtest/regtest.sh"] + args, stdout=subprocess.PIPE, universal_newlines=True)
    process.wait(timeout=timeout)
    assert process.returncode == 0
    return process.stdout.read().strip()


def data_check(data, key, check_type, length=None):
    assert key in data
    assert isinstance(data[key], check_type)
    if length:
        assert len(data[key]) == length



from .bch import BCH
from .bsty import BSTY
from .btc import BTC
from .gzro import GZRO
from .ltc import LTC
from .xrg import XRG

COINS = {"BTC": BTC, "BCH": BCH, "LTC": LTC, "GZRO": GZRO, "BSTY": BSTY, "XRG": XRG}

__all__ = list(COINS.keys()) + ["COINS"]



import asyncio
from typing import Any, Callable, Optional
from urllib.parse import urljoin

import aiohttp
from jsonrpcclient.clients.aiohttp_client import AiohttpClient as RPC
from jsonrpcclient.exceptions import ReceivedErrorResponseError
from jsonrpcclient.requests import Request

from ..errors import ConnectionFailedError, UnknownError, generate_exception
from ..utils import json_encode


class RPCProxy:
    session: aiohttp.ClientSession

    def __init__(
        self,
        url: str,
        username: Optional[str] = None,
        password: Optional[str] = None,
        xpub: Optional[str] = None,
        session: Optional[aiohttp.ClientSession] = None,
        proxy: Optional[str] = None,
        verify: Optional[bool] = True,
    ):
        self.url = url
        self.username = username
        self.password = password
        self.xpub = xpub
        self.proxy = proxy
        self.verify = verify
        self._connector_class = aiohttp.TCPConnector
        self._connector_init = dict(ssl=self.verify)
        self._loop = asyncio.get_event_loop()
        self._loop.set_exception_handler(lambda loop, context: None)
        self._spec = {"exceptions": {"-32600": {"exc_name": "UnauthorizedError", "docstring": "Unauthorized"}}}
        self._spec_valid = False
        if session:
            self.sesson = session
        else:
            self.create_session()
        self.rpc = RPC(endpoint=self.url, session=self.session, timeout=5 * 60)

    def init_proxy(self) -> None:
        if self.proxy:
            from aiohttp_socks import ProxyConnector
            from aiohttp_socks.utils import parse_proxy_url

            proxy_type, host, port, username, password = parse_proxy_url(self.proxy)
            self._connector_class = ProxyConnector
            self._connector_init.update(
                proxy_type=proxy_type,
                host=host,
                port=port,
                username=username,
                password=password,
                rdns=True,
            )

    def create_session(self) -> None:
        self.init_proxy()
        self.session = aiohttp.ClientSession(
            connector=self._connector_class(**self._connector_init),  # type: ignore
            loop=self._loop,
            auth=aiohttp.BasicAuth(self.username, self.password),  # type: ignore
        )

    def validate_spec(self, spec: Any) -> bool:
        if not isinstance(spec, dict):
            return False
        if not spec.keys() >= {"version", "exceptions"}:
            return False
        if not isinstance(spec["version"], str) or not isinstance(spec["exceptions"], dict):
            return False
        if not all(
            isinstance(code, str) and isinstance(exc, dict) and exc.keys() >= {"exc_name", "docstring"}
            for code, exc in spec["exceptions"].items()
        ):
            return False
        return True

    async def fetch_spec(self) -> Any:
        resp = await self.session.get(urljoin(self.url, "/spec"))
        spec = await resp.json()
        return spec

    @property
    async def spec(self) -> dict:
        if self._spec_valid:
            return self._spec
        try:
            spec = await self.fetch_spec()
            self._spec_valid = self.validate_spec(spec)
            if self._spec_valid:
                self._spec = spec
        except Exception:
            pass
        return self._spec

    def __getattr__(self, method: str, *args: Any, **kwargs: Any) -> Callable:
        from ..sync import async_to_sync_wraps

        @async_to_sync_wraps
        async def wrapper(*args: Any, **kwargs: Any) -> Any:
            try:
                return (
                    await self.rpc.send(
                        json_encode(Request(method, xpub=self.xpub, *args, **kwargs)),
                        validate_against_schema=False,
                    )
                ).data.result
            except ReceivedErrorResponseError as e:
                message = e.response.message
                error_code = str(e.response.code)
                exceptions = (await self.spec)["exceptions"]
                if error_code in exceptions:
                    exc = exceptions[error_code]
                    raise generate_exception(exc["exc_name"])(exc["docstring"]) from e
                raise UnknownError("Unknown error from server: {}".format(message)) from e
            except aiohttp.ClientConnectionError as e:
                raise ConnectionFailedError() from e

        return wrapper

    async def _close(self) -> None:
        await self.session.close()

    def __del__(self) -> None:
        if self._loop.is_running():
            self._loop.create_task(self._close())
        else:
            self.session._connector._closed = True  # type: ignore



import socket

import pytest
from aiohttp import ClientSession, TCPConnector, web
from aiohttp.resolver import DefaultResolver
from aiohttp.test_utils import unused_port

from bitcart import BTC
from bitcart.coin import Coin

TEST_XPUB = "zprvAWgYBBk7JR8GkUwcwyhT1qk8FQpym8cHboGyDEdMhHcL1NRe8bioZR3uo5gTSTG47iqQX6VqPL6iHHDNK55taJiV9MEscu6UiqSR1tAiSUq"
REGTEST_XPUB = (
    "vprv9DMUxX4ShgxMMQduKMSnoNsX4jZjdqmRRapGRRFbok1XXrjkvFyAnvSVPbs4t8ZDA73fTT9DLzdCyHBh39AZHG8nsP1gEj11EwSdYP8zhKF"
)
LIGHTNING_UNSUPPORTED_XPUB = (
    "xprv9s21ZrQH143K3tZPHG8CbfZ7uUY5stdHmaEXeSqawGrZuAoBdHPgKHjdkfmHSdxDJSbo29JiU1PcWhzEsgFryqMHQfr2T5TWBPK8EqFjscZ"
)


@pytest.fixture
async def xpub():
    return TEST_XPUB


@pytest.fixture
async def coin():
    return Coin()


@pytest.fixture
async def btc_nowallet():
    return BTC()


@pytest.fixture
async def btc_wallet():
    return BTC(xpub=TEST_XPUB)


@pytest.fixture
async def lightning_unsupported_wallet():
    return BTC(xpub=LIGHTNING_UNSUPPORTED_XPUB)


@pytest.fixture
async def regtest_wallet():
    return BTC(xpub=REGTEST_XPUB)


@pytest.fixture
async def regtest_node_id():
    return await BTC(xpub=REGTEST_XPUB, rpc_url="http://localhost:5110").node_id


@pytest.fixture
async def btc():
    return BTC()


class FakeResolver:
    _LOCAL_HOST = {0: "127.0.0.1", socket.AF_INET: "127.0.0.1", socket.AF_INET6: "::1"}

    def __init__(self, fakes):
        """fakes -- dns -> port dict"""
        self._fakes = fakes
        self._resolver = DefaultResolver()

    async def resolve(self, host, port=0, family=socket.AF_INET):
        fake_port = self._fakes.get(host)
        if fake_port is not None:
            return [
                {
                    "hostname": host,
                    "host": self._LOCAL_HOST[family],
                    "port": fake_port,
                    "family": family,
                    "proto": 0,
                    "flags": socket.AI_NUMERICHOST,
                }
            ]
        else:
            return await self._resolver.resolve(host, port, family)


class FakeDaemon:
    def __init__(self):
        self.app = web.Application()
        self.app.router.add_routes([web.get("/ws", self.handle_websocket)])
        self.runner = None

    async def start(self):
        port = unused_port()
        self.runner = web.AppRunner(self.app)
        await self.runner.setup()
        site = web.TCPSite(self.runner, "127.0.0.1", port)
        await site.start()
        return {"localhost": port}

    async def stop(self):
        await self.runner.cleanup()

    async def handle_websocket(self, request):
        ws = web.WebSocketResponse()
        await ws.prepare(request)
        await self.reply_to_websocket(ws)
        await ws.close()

    async def reply_to_websocket(self, ws):
        await ws.send_json({"wallet": TEST_XPUB, "updates": [{"event": "new_transaction", "tx": "test"}]})


class FakeBadJSONDaemon(FakeDaemon):
    async def reply_to_websocket(self, ws):
        await ws.send_str("")


async def patched_session_maker(daemon_class):
    fake_daemon = daemon_class()
    info = await fake_daemon.start()
    resolver = FakeResolver(info)
    connector = TCPConnector(resolver=resolver)
    fake_session = ClientSession(connector=connector)
    return fake_session, fake_daemon


@pytest.fixture
async def patched_session():
    session, daemon = await patched_session_maker(FakeDaemon)
    yield session
    await daemon.stop()


@pytest.fixture
async def patched_session_bad_json():
    session, daemon = await patched_session_maker(FakeBadJSONDaemon)
    yield session
    await daemon.stop()



import multiprocessing

import pytest

from bitcart.errors import ConnectionFailedError

test_queue = multiprocessing.Queue()

pytestmark = pytest.mark.asyncio


def new_tx_handler(event, tx):
    test_queue.put(True)


async def test_event_delivery(patched_session, btc_wallet, mocker):
    btc_wallet.add_event_handler("new_transaction", new_tx_handler)
    mocker.patch.object(btc_wallet.server, "session", patched_session)
    await btc_wallet.start_websocket(auto_reconnect=False)
    assert test_queue.qsize() == 1
    assert test_queue.get() is True


async def test_bad_connection(btc_wallet):
    btc_wallet.server.url = "http://localhost1:11234"  # nothing running
    with pytest.raises(ConnectionFailedError):
        await btc_wallet.start_websocket(auto_reconnect=False)


async def test_bad_json(patched_session_bad_json, btc_wallet, mocker):
    btc_wallet.add_event_handler("new_transaction", new_tx_handler)
    mocker.patch.object(btc_wallet.server, "session", patched_session_bad_json)
    await btc_wallet.start_websocket(auto_reconnect=False)
    assert test_queue.qsize() == 0


async def test_bad_json_reconnect_callback(patched_session_bad_json, btc_wallet, mocker):
    btc_wallet.add_event_handler("new_transaction", new_tx_handler)
    mocker.patch.object(btc_wallet.server, "session", patched_session_bad_json)
    await btc_wallet.start_websocket(auto_reconnect=False, reconnect_callback=lambda: test_queue.put(True))
    assert test_queue.qsize() == 1  # from reconnect_callback
    assert test_queue.get() is True



from decimal import Decimal

import pytest

from bitcart.utils import bitcoins, convert_amount_type, satoshis


@pytest.mark.parametrize("btc,expected", [(0.1, 10000000), (1, 100000000), (0.00000001, 1), (5, 500000000)])
def test_satoshis(btc, expected):
    result = satoshis(btc)
    assert isinstance(result, int)
    assert result == expected


@pytest.mark.parametrize("sats,expected", [(10000000, "0.1"), (100000000, "1"), (1, "0.00000001"), (500000000, "5")])
def test_bitcoins(sats, expected):
    result = bitcoins(sats)
    assert isinstance(result, Decimal)
    assert result == Decimal(expected)


def test_convertability():
    assert bitcoins(satoshis(1)) == 1


def test_convert_amount_type():
    assert convert_amount_type("1") == Decimal("1")


@pytest.mark.asyncio
async def test_decimal_sending(btc_wallet):
    amount = Decimal("0.5")
    req = await btc_wallet.add_request(amount)  # ensures that it is possible to pass decimal
    assert req[btc_wallet.amount_field] == amount



from decimal import Decimal

import pytest

from bitcart import errors
from bitcart.errors import LightningDisabledError

from ...utils import data_check

pytestmark = pytest.mark.asyncio


async def test_node_id(btc_wallet):
    assert isinstance(await btc_wallet.node_id, str)


async def test_add_invoice(btc_wallet):
    invoice = await btc_wallet.add_invoice(0.5, "test description")
    assert isinstance(invoice, dict)
    assert (
        invoice.items()
        > {
            "is_lightning": True,
            "amount_BTC": Decimal("0.5"),
            "message": "test description",
            "expiration": 900,
            "status": 0,
            "status_str": "Expires in 15 minutes",
            "amount_msat": 50000000000,
            "can_receive": False,
        }.items()
    )
    data_check(invoice, "timestamp", int)
    data_check(invoice, "rhash", str)
    data_check(invoice, "invoice", str)
    got_invoice = await btc_wallet.get_invoice(invoice["rhash"])
    assert got_invoice == invoice


async def test_connect(btc_wallet):
    with pytest.raises(errors.InvalidNodeIDError):
        assert not await btc_wallet.connect("")
    with pytest.raises(errors.InvalidNodeIDError):
        assert not await btc_wallet.connect("172.81.181.3")  # taken from electrum sources
    assert await btc_wallet.connect("0214382bdce7750dfcb8126df8e2b12de38536902dc36abcebdaeefdeca1df8284@172.81.181.3")


async def test_list_peers(btc_wallet):
    res1 = await btc_wallet.list_peers()
    res2 = await btc_wallet.list_peers(True)
    assert isinstance(res1, list)
    assert isinstance(res2, list)
    assert len(res1) > 0
    assert len(res2) >= 0
    peer = res1[0]
    assert peer.keys() == {"node_id", "address", "initialized", "features", "channels"}
    data_check(peer, "initialized", bool)
    data_check(peer, "node_id", str, 66)
    data_check(peer, "address", str)
    data_check(peer, "features", str)
    assert "LnFeatures." in peer["features"]
    data_check(peer, "channels", list, 0)


async def test_lightning_disabled(btc_wallet):
    await btc_wallet.set_config("lightning", False)
    with pytest.raises(LightningDisabledError):
        await btc_wallet.node_id
    await btc_wallet.set_config("lightning", True)  # reset back


async def test_wallet_methods_on_non_segwit(lightning_unsupported_wallet):
    with pytest.raises(errors.LightningUnsupportedError):
        await lightning_unsupported_wallet.list_channels()  # unsupported on non-segwit wallets



import multiprocessing

import pytest

from bitcart import BTC, GZRO, LTC
from bitcart.errors import NoCurrenciesRegisteredError
from bitcart.manager import APIManager

pytestmark = pytest.mark.asyncio

test_queue = multiprocessing.Queue()
test_queue2 = multiprocessing.Queue()


class DummyServer:
    url = "http://localhost:5000"
    session = None


def new_tx_handler(instance, event, tx):
    if isinstance(instance, BTC):
        test_queue.put(True)


def reconnect_callback(currency):
    test_queue2.put(currency)


@pytest.fixture
async def manager(xpub):
    return APIManager({"BTC": [xpub, xpub]})


@pytest.fixture
async def websocket_manager(xpub):
    return APIManager({"BTC": [xpub, "test"]})


@pytest.fixture
async def websocket_manager_no_wallets():
    return APIManager({"BTC": []})


async def test_manager_storage(manager, xpub):
    assert manager.wallets == {"BTC": {xpub: BTC(xpub=xpub)}}
    assert manager.wallets.BTC == manager.wallets["BTC"] == manager.BTC == manager["BTC"]


async def test_manager_classmethods(xpub):
    assert APIManager.load_wallet("BTC", xpub) == BTC(xpub=xpub)
    assert APIManager.load_wallets("BTC", [xpub, xpub]) == {xpub: BTC(xpub=xpub)}


async def test_manager_add_wallets(manager, xpub):
    assert manager.add_wallet("LTC", xpub) is None
    assert manager.wallets == {"BTC": {xpub: BTC(xpub=xpub)}, "LTC": {xpub: LTC(xpub=xpub)}}
    assert manager.add_wallets("GZRO", [xpub, xpub]) is None
    assert manager.wallets == {"BTC": {xpub: BTC(xpub=xpub)}, "LTC": {xpub: LTC(xpub=xpub)}, "GZRO": {xpub: GZRO(xpub=xpub)}}


async def test_manager_start_websocket(patched_session, websocket_manager, xpub, mocker):
    websocket_manager.add_event_handler("new_transaction", new_tx_handler)
    mocker.patch.object(websocket_manager.BTC[xpub].server, "session", patched_session)
    await websocket_manager.start_websocket(auto_reconnect=False)
    assert test_queue.qsize() == 1
    assert test_queue.get() is True


async def test_manager_reconnect_callback(patched_session, websocket_manager, xpub, mocker):
    mocker.patch.object(websocket_manager.BTC[xpub].server, "session", patched_session)
    await websocket_manager.start_websocket(auto_reconnect=False, reconnect_callback=reconnect_callback)
    assert test_queue2.qsize() == 1
    assert test_queue2.get() == "BTC"


async def test_manager_no_wallets(patched_session, websocket_manager_no_wallets, mocker):
    websocket_manager_no_wallets.add_event_handler("new_transaction", new_tx_handler)
    server = DummyServer()
    server.session = patched_session
    mocker.patch("bitcart.APIManager._get_websocket_server", return_value=server)
    await websocket_manager_no_wallets.start_websocket(auto_reconnect=False)
    assert test_queue.qsize() == 1
    assert test_queue.get() is True


async def test_manager_no_currencies():
    with pytest.raises(NoCurrenciesRegisteredError):
        await APIManager().start_websocket(auto_reconnect=False)



from decimal import Decimal
from functools import wraps
from typing import TYPE_CHECKING, Any, Callable, Dict, Iterable, Optional, Union

from ..coin import Coin
from ..errors import InvalidEventError, LightningDisabledError
from ..event_delivery import EventDelivery
from ..providers.jsonrpcrequests import RPCProxy
from ..types import AmountType
from ..utils import bitcoins, call_universal, convert_amount_type

if TYPE_CHECKING:
    from aiohttp import ClientSession, ClientWebSocketResponse


def lightning(f: Callable) -> Callable:
    @wraps(f)
    async def wrapper(*args: Any, **kwargs: Any) -> Any:
        if len(args) > 0:
            obj = args[0]
            if not await obj.get_config("lightning"):
                raise LightningDisabledError("Lightning is disabled in current daemon.")
        return await f(*args, **kwargs)

    return wrapper


class BTC(Coin, EventDelivery):
    coin_name = "BTC"
    friendly_name = "Bitcoin"
    RPC_URL = "http://localhost:5000"
    RPC_USER = "electrum"
    RPC_PASS = "electrumz"
    ALLOWED_EVENTS = ["new_block", "new_transaction", "new_payment"]
    BALANCE_ATTRS = ["confirmed", "unconfirmed", "unmatured", "lightning"]

    def __init__(
        self,
        rpc_url: Optional[str] = None,
        rpc_user: Optional[str] = None,
        rpc_pass: Optional[str] = None,
        xpub: Optional[str] = None,
        proxy: Optional[str] = None,
        session: Optional["ClientSession"] = None,
    ):
        super().__init__()
        self.rpc_url = rpc_url or self.RPC_URL
        self.rpc_user = rpc_user or self.RPC_USER
        self.rpc_pass = rpc_pass or self.RPC_PASS
        self.xpub = xpub
        self.event_handlers: Dict[str, Callable] = {}
        self.amount_field = getattr(self, "AMOUNT_FIELD", f"amount_{self.coin_name}")
        self.server = RPCProxy(self.rpc_url, self.rpc_user, self.rpc_pass, self.xpub, session=session, proxy=proxy)

    @property
    async def spec(self) -> dict:
        """Returns current daemon's spec

        It contains documentation for all possible exceptions raised

        Example:

        >>> c.spec
        {'version': '0.0.1', 'electrum_map': {...}, 'exceptions': {...}}

        Returns:
            dict: spec
        """
        return await self.server.spec

    ### High level interface ###

    async def help(self) -> list:
        return await self.server.help()  # type: ignore

    async def get_tx(self, tx: str) -> dict:
        return await self.server.get_transaction(tx)  # type: ignore

    async def get_address(self, address: str) -> list:
        out: list = await self.server.getaddresshistory(address)
        for i in out:
            i["tx"] = await self.get_tx(i["tx_hash"])
        return out

    async def balance(self) -> dict:
        data = await self.server.getbalance()
        return {attr: convert_amount_type(data.get(attr, 0)) for attr in self.BALANCE_ATTRS}

    async def _add_request(self, *args: Any, **kwargs: Any) -> dict:
        return await self.server.add_request(*args, **kwargs)  # type: ignore

    def _convert_amounts(self, data: dict) -> dict:
        if data[self.amount_field].lower() != "unknown":
            data[self.amount_field] = convert_amount_type(data[self.amount_field])
        return data

    async def _add_request_base(
        self,
        method: Callable,
        amount: Optional[AmountType] = None,
        description: str = "",
        expire: Union[int, float] = 15,
        extra_kwargs: dict = {},
    ) -> dict:
        expiration = 60 * expire if expire else None
        kwargs = {"amount": amount, "memo": description, "expiration": expiration}
        kwargs.update(extra_kwargs)
        data = await method(**kwargs)
        return self._convert_amounts(data)

    async def add_request(
        self,
        amount: Optional[AmountType] = None,
        description: str = "",
        expire: Union[int, float] = 15,
    ) -> dict:
        """Add invoice

        Create an invoice and request amount in BTC, it will expire by parameter provided.
        If expire is None, it will last forever.

        Example:

        >>> c.add_request(0.5, "My invoice", 20)
        {'time': 1562762334, 'amount': 50000000, 'exp': 1200, 'address': 'xxx',...

        Args:
            self (BTC): self
            amount (Optional[AmountType]): amount to open invoice. Defaults to None.
            description (str, optional): Description of invoice. Defaults to "".
            expire (Union[int, float], optional): The time invoice will expire in. In minutes. Defaults to 15.

        Returns:
            dict: Invoice data
        """
        return await self._add_request_base(self._add_request, amount, description, expire, extra_kwargs={"force": True})

    async def get_request(self, address: str) -> dict:
        """Get invoice info

        Get invoice information by address got from add_request

        Example:

        >>> c.get_request("1A6jnc6xQwmhsChNLcyKAQNWPcWsVYqCqJ")
        {'time': 1562762334, 'amount': 50000000, 'exp': 1200, 'address': '1A6jnc6xQwmhsChNLcyKAQNWPcWsVYqCqJ',...

        Args:
            self (BTC): self
            address (str): address of invoice

        Returns:
            dict: Invoice data
        """
        data = await self.server.getrequest(address)
        return self._convert_amounts(data)

    async def history(self) -> dict:
        """Get transaction history of wallet

        Example:

        >>> c.history()
        {'summary': {'end_balance': '0.', 'end_date': None, 'from_height': None, 'incoming': '0.00185511',...

        Args:
            self (BTC): self

        Returns:
            dict: dictionary with some data, where key transactions is list of transactions
        """
        return await self.server.onchain_history()  # type: ignore

    async def _register_wallets(self, ws: "ClientWebSocketResponse") -> None:
        await ws.send_json({"xpub": self.xpub})

    async def process_updates(self, updates: Iterable[dict], *args: Any, pass_instance: bool = False, **kwargs: Any) -> None:
        if not isinstance(updates, list):
            return
        for event_info in updates:
            if not isinstance(event_info, dict):
                continue
            event = event_info.pop("event", None)
            if not event or event not in self.ALLOWED_EVENTS:
                raise InvalidEventError(f"Invalid event from server: {event}")
            handler = self.event_handlers.get(event)
            if handler:
                args = (event,)
                if pass_instance:
                    args = (self,) + args
                try:
                    await call_universal(handler, *args, **event_info)
                except Exception:
                    pass

    async def pay_to(
        self,
        address: str,
        amount: AmountType,
        fee: Optional[Union[AmountType, Callable]] = None,
        feerate: Optional[AmountType] = None,
        broadcast: bool = True,
    ) -> Union[dict, str]:
        """Pay to address

        This function creates a transaction, your wallet must have sufficent balance
        and address must exist.

        Examples:

        >>> btc.pay_to("mkHS9ne12qx9pS9VojpwU5xtRd4T7X7ZUt", 0.001)
        '608d9af34032868fd2849723a4de9ccd874a51544a7fba879a18c847e37e577b'

        >>> btc.pay_to("mkHS9ne12qx9pS9VojpwU5xtRd4T7X7ZUt",0.001, feerate=1)
        '23d0aec06f6ea6100ba9c6ce8a1fa5d333a6c1d39a780b5fadc4b2836d71b66f'

        >>> btc.pay_to("mkHS9ne12qx9pS9VojpwU5xtRd4T7X7ZUt", 0.001, broadcast=False)
        {'hex': '02000000026.....', 'complete': True, 'final': False, 'name': None, 'csv_delay': 0, 'cltv_expiry': 0}

        Args:
            self (BTC): self
            address (str): address where to send BTC
            amount (AmountType): amount of bitcoins to send
            fee (Optional[Union[AmountType, Callable]], optional): Either a fixed fee, or a callable getting size and
                default fee as argument and returning fee. Defaults to None.
            feerate (Optional[AmountType], optional): A sat/byte feerate, can't be passed together with fee argument.
                Defaults to None.
            broadcast (bool, optional): Whether to broadcast transaction to network. Defaults to True.

        Raises:
            TypeError: if you have provided both fee and feerate

        Returns:
            Union[dict, str]: tx hash of ready transaction or raw transaction, depending on broadcast argument.
        """
        if fee and feerate:
            raise TypeError("Can't specify both fee and feerate at the same time")
        is_callable = callable(fee)
        fee_arg = fee if not is_callable else None
        tx_data = await self.server.payto(
            address, amount, fee=fee_arg, feerate=feerate, addtransaction=broadcast and not is_callable
        )
        if is_callable:
            tx_size = await self.server.get_tx_size(tx_data)
            default_fee = await self.server.get_default_fee(tx_size)
            try:
                resulting_fee: Optional[str] = str(bitcoins(fee(tx_size, default_fee)))  # type: ignore
            except Exception:
                resulting_fee = None
            if resulting_fee:
                tx_data = await self.server.payto(
                    address, amount, fee=resulting_fee, feerate=feerate, addtransaction=broadcast
                )
            elif broadcast:  # use existing tx_data
                await self.server.addtransaction(tx_data)
        if broadcast:
            return await self.server.broadcast(tx_data)  # type: ignore
        else:
            return tx_data  # type: ignore

    async def pay_to_many(
        self,
        outputs: Iterable[Union[dict, tuple]],
        fee: Optional[Union[AmountType, Callable]] = None,
        feerate: Optional[AmountType] = None,
        broadcast: bool = True,
    ) -> Union[dict, str]:
        """Pay to multiple addresses(batch transaction)

        This function creates a batch transaction, your wallet must have sufficent balance
        and addresses must exist.
        outputs parameter is either an iterable of ``(address, amount)`` tuples(or any iterables) or a dict with two
        keys: address and amount ``{"address": "someaddress", "amount": 0.5}``

        Examples:

        >>> btc.pay_to_many([{"address":"mkHS9ne12qx9pS9VojpwU5xtRd4T7X7ZUt","amount":0.001}, \
{"address":"mv4rnyY3Su5gjcDNzbMLKBQkBicCtHUtFB","amount":0.0001}])
        '60fa120d9f868a7bd03d6bbd1e225923cab0ba7a3a6b961861053c90365ed40a'

        >>> btc.pay_to_many([("mkHS9ne12qx9pS9VojpwU5xtRd4T7X7ZUt",0.001), ("mv4rnyY3Su5gjcDNzbMLKBQkBicCtHUtFB",0.0001)])
        'd80f14e20af2ceaa43a8b7e15402d420246d39e235d87874f929977fb0b1cab8'

        >>> btc.pay_to_many((("mkHS9ne12qx9pS9VojpwU5xtRd4T7X7ZUt",0.001), \
("mkHS9ne12qx9pS9VojpwU5xtRd4T7X7ZUt",0.001)), feerate=1)
        '0a6611876e04a6f2742eac02d4fac4c242dda154d85f0d547bbac1a33dbbbe34'

        >>> btc.pay_to_many([("mkHS9ne12qx9pS9VojpwU5xtRd4T7X7ZUt",0.001), \
("mv4rnyY3Su5gjcDNzbMLKBQkBicCtHUtFB",0.0001)], broadcast=False)
        {'hex': '0200000...', 'complete': True, 'final': False}

        Args:
            self (BTC): self
            outputs (Iterable[Union[dict, tuple]]): An iterable with dictionary or iterable as the item
            fee (Optional[Union[AmountType, Callable]], optional): Either a fixed fee, or a callable getting size and
                default fee as argument and returning fee. Defaults to None.
            feerate (Optional[AmountType], optional): A sat/byte feerate, can't be passed together with fee argument.
                Defaults to None.
            broadcast (bool, optional): Whether to broadcast transaction to network. Defaults to True.

        Raises:
            TypeError: if you have provided both fee and feerate

        Returns:
            Union[dict, str]: tx hash of ready transaction or raw transaction, depending on broadcast argument.
        """
        if fee and feerate:
            raise TypeError("Can't specify both fee and feerate at the same time")
        new_outputs = []
        dict_outputs = False
        for output in outputs:
            if isinstance(output, dict):
                dict_outputs = True
                new_outputs.append((output["address"], output["amount"]))
        outputs = new_outputs if dict_outputs else outputs
        is_callable = callable(fee)
        fee_arg = fee if not is_callable else None
        tx_data = await self.server.paytomany(
            outputs, fee=fee_arg, feerate=feerate, addtransaction=broadcast and not is_callable
        )
        if is_callable:
            tx_size = await self.server.get_tx_size(tx_data)
            default_fee = await self.server.get_default_fee(tx_size)
            try:
                resulting_fee: Optional[str] = str(bitcoins(fee(tx_size, default_fee)))  # type: ignore
            except Exception:
                resulting_fee = None
            if resulting_fee:
                tx_data = await self.server.paytomany(outputs, fee=resulting_fee, feerate=feerate, addtransaction=broadcast)
            elif broadcast:  # use existing tx_data
                await self.server.addtransaction(tx_data)
        if broadcast:
            return await self.server.broadcast(tx_data)  # type: ignore
        else:
            return tx_data  # type: ignore

    async def rate(self, currency: str = "USD") -> Decimal:
        """Get bitcoin price in selected fiat currency

        It uses the same method as electrum wallet gets exchange rate-via different payment providers

        Examples:

        >>> c.rate()
        9878.527

        >>> c.rate("RUB")
        757108.226

        Args:
            self (BTC): self
            currency (str, optional): Currency to get rate into. Defaults to "USD".

        Returns:
            Decimal: price of 1 bitcoin in selected fiat currency
        """
        return convert_amount_type(await self.server.exchange_rate(currency))

    async def list_fiat(self) -> Iterable[str]:
        """List of all available fiat currencies to get price for

        This list is list of only valid currencies that could be passed to rate() function

        Example:

        >>> c.list_fiat()
        ['AED', 'ARS', 'AUD', 'BCH', 'BDT', 'BHD', 'BMD', 'BNB', 'BRL', 'BTC', ...]

        Args:
            self (BTC): self

        Returns:
            Iterable[str]: list of available fiat currencies
        """
        return await self.server.list_currencies()  # type: ignore

    async def set_config(self, key: str, value: Any) -> bool:
        """Set config key to specified value

        It sets the config value in electrum's config store, usually
        $HOME/.electrum/config

        You can set any keys and values using this function(as long as JSON serializable),
        and some are used to configure underlying electrum daemon.

        Example:

        >>> c.set_config("x", 5)
        True

        Args:
            self (BTC): self
            key (str): key to set
            value (Any): value to set

        Returns:
            bool: True on success, False otherwise
        """
        return await self.server.setconfig(key, value)  # type: ignore

    async def get_config(self, key: str, default: Any = None) -> Any:
        """Get config key

        If the key doesn't exist, default value is returned.
        Keys are stored in electrum's config file, check :meth:`bitcart.coins.btc.BTC.set_config` doc for details.

        Example:

        >>> c.get_config("x")
        5

        Args:
            self (BTC): self
            key (str): key to get
            default (Any, optional): The value to default to when key doesn't exist. Defaults to None.

        Returns:
            Any: value of the key or default value provided
        """
        return await self.server.getconfig(key) or default

    async def validate_key(self, key: str) -> bool:
        """Validate whether provided key is valid to restore a wallet

        If the key is x/y/z pub/prv or electrum seed at the network daemon is running
        at, then it would be valid(True), else False

        Examples:

        >>> c.validate_key("test")
        False

        >>> c.validate_key("your awesome electrum seed")
        True

        >>> c.validate_key("x/y/z pub/prv here")
        True

        Args:
            self (BTC): self
            key (str): key to check

        Returns:
            bool: Whether the key is valid or not
        """
        return await self.server.validatekey(key)  # type: ignore

    ### Lightning apis ###

    @lightning
    async def open_channel(self, node_id: str, amount: AmountType) -> str:
        """Open lightning channel

        Open channel with node, returns string of format
        txid:output_index

        Args:
            self (BTC): self
            node_id (str): id of node to open channel with
            amount (AmountType): amount to open channel

        Returns:
            str: string of format txid:output_index
        """
        return await self.server.open_channel(node_id, amount)  # type: ignore

    @lightning
    async def add_invoice(
        self,
        amount: AmountType,
        description: str = "",
        expire: Union[int, float] = 15,
    ) -> dict:
        """Create a lightning invoice

        Create a lightning invoice and return invoice data with bolt invoice id
        All parameters are the same as in add_request

        Example:

        >>> a.add_invoice(0.5, "My invoice", 20)
        {'time': 1562762334, 'amount': 50000000, 'exp': 1200, 'invoice': 'lnbc500m',...

        Args:
            self (BTC): self
            amount (AmountType): amount to open invoice
            description (str, optional): Description of invoice. Defaults to "".
            expire (Union[int, float], optional): The time invoice will expire in. In minutes. Defaults to 15.

        Returns:
            dict: Invoice data
        """
        return await self._add_request_base(self.server.add_lightning_request, amount, description, expire)

    @lightning
    async def close_channel(self, channel_id: str, force: bool = False) -> str:
        """Close lightning channel

        Close channel by channel_id got from open_channel, returns transaction id

        Args:
            self (BTC): self
            channel_id (str): channel_id from open_channel
            force (bool): Create new address beyond gap limit, if no more addresses are available.

        Returns:
            str: tx_id of closed channel
        """
        return await self.server.close_channel(channel_id, force)  # type: ignore

    @property  # type: ignore
    @lightning
    async def node_id(self) -> str:
        """Get node id

        Electrum's lightning implementation itself is a lightning node,
        that way you can get a super light node, this method returns it's id

        Example:

        >>> a.node_id
        '030ff29580149a366bdddf148713fa808f0f4b934dccd5f7820f3d613e03c86e55'

        Returns:
            str: id of your node
        """
        return await self.server.nodeid()  # type: ignore

    @lightning
    async def lnpay(self, invoice: str) -> bool:
        """Pay lightning invoice

        Returns True on success, False otherwise

        Args:
            invoice (str): invoice to pay

        Returns:
            bool: success or not
        """
        return await self.server.lnpay(invoice)  # type: ignore

    @lightning
    async def connect(self, connection_string: str) -> bool:
        """Connect to lightning node

        connection string must respect format pubkey@ipaddress

        Args:
            connection_string (str): connection string

        Returns:
            bool: True on success, False otherwise
        """
        return await self.server.add_peer(connection_string)  # type: ignore

    @lightning
    async def list_peers(self, gossip: bool = False) -> list:
        """Get a list of lightning peers

        Args:
            gossip (bool, optional): Whether to return peers of a gossip (one per node) or of a wallet. Defaults to False.

        Returns:
            list: list of lightning peers
        """
        return await self.server.list_peers(gossip=gossip)  # type: ignore

    @lightning
    async def list_channels(self) -> list:
        """List all channels ever opened

        Possible channel statuses:
        OPENING, OPEN, CLOSED, DISCONNECTED

        Example:

        >>> a.server.list_channels()
        [{'local_htlcs': {'adds': {}, 'locked_in': {}, 'settles': {}, 'fails': {}}, 'remote_htlcs': ...

        Returns:
            list: list of channels
        """
        return await self.server.list_channels()  # type: ignore

    @lightning
    async def get_invoice(self, rhash: str) -> dict:
        """Get lightning invoice info

        Get lightning invoice information by rhash got from add_invoice

        Example:

        >>> c.get_invoice("e34d7fb4cda66e0760fc193496c302055d0fd960cfd982432355c8bfeecd5f33")
        {'is_lightning': True, 'amount_BTC': Decimal('0.5'), 'timestamp': 1619273042, 'expiration': 900, ...

        Args:
            rhash (str): invoice rhash

        Returns:
            dict: invoice data
        """
        data = await self.server.get_invoice(rhash)
        return self._convert_amounts(data)



from typing import Callable, Dict, Optional


class Coin:
    """Coins should reimplement some methods,
    and initialize coin-specific info.
    Required information is:
    coin_name str
    friendly_name str
    For more info see the docs.
    """

    coin_name = "Base"
    friendly_name = "Base"
    event_handlers: Dict[str, Callable]
    xpub: Optional[str]

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Coin):
            return NotImplemented
        return (self.coin_name, self.xpub) == (other.coin_name, other.xpub)

    async def help(self) -> list:
        """Get help

        Returns a list of all available RPC methods

        Raises:
            NotImplementedError: Implement in your subclass

        Returns:
            list: RPC methods list
        """
        raise NotImplementedError()

    async def get_tx(self, tx: str) -> dict:
        """Get transaction information

        Given tx hash of transaction, return full information as dictionary

        Example:

        >>> c.get_tx("54604b116b28124e31d2d20bbd4561e6f8398dca4b892080bffc8c87c27762ba")
        {'partial': False, 'version': 2, 'segwit_ser': True, 'inputs': [{'prevout_hash': 'xxxx',...

        Args:
            tx (str): tx_hash

        Raises:
            NotImplementedError: Implement in your subclass

        Returns:
            dict: transaction info
        """
        raise NotImplementedError()

    async def get_address(self, address: str) -> list:
        """Get address history

        This method should return list of transaction informations for specified address

        Example:

        >>> c.get_address("31smpLFzLnza6k8tJbVpxXiatGjiEQDmzc")
        [{'tx_hash': '7854bdf4c4e27276ecc1fb8d666d6799a248f5e81bdd58b16432d1ddd1d4c332', 'height': 581878, 'tx': ...

        Args:
            address (str): address to get transactions for

        Raises:
            NotImplementedError: Override this method in subclass

        Returns:
            list: List of transactions
        """
        raise NotImplementedError()

    async def balance(self) -> dict:
        """Get balance of wallet

        Example:

        >>> self.balance()
        {"confirmed": 0.00005, "unconfirmed": 0, "unmatured": 0}

        Raises:
            NotImplementedError: Implement in your subclass

        Returns:
            dict: It should return dict of balance statuses
        """
        raise NotImplementedError()



# Thanks to https://github.com/pyrogram/pyrogram/blob/master/pyrogram/sync.py for most of the implementation
import asyncio
import functools
import inspect
import threading
from typing import Any, AsyncGenerator, Callable, List

from .coins import COINS
from .manager import APIManager
from .providers.jsonrpcrequests import RPCProxy


async def consume_generator(coroutine: AsyncGenerator) -> List[Any]:
    return [i async for i in coroutine]


def run_sync_ctx(coroutine: Any, loop: asyncio.AbstractEventLoop) -> Any:
    if inspect.iscoroutine(coroutine):
        return loop.run_until_complete(coroutine)

    if inspect.isasyncgen(coroutine):
        return loop.run_until_complete(consume_generator(coroutine))

    return coroutine


def async_to_sync_wraps(function: Callable, is_property: bool = False) -> Callable:
    main_loop = asyncio.get_event_loop()

    @functools.wraps(function)
    def async_to_sync_wrap(*args: Any, **kwargs: Any) -> Any:
        try:
            loop = asyncio.get_event_loop()
        except RuntimeError:
            loop = main_loop
        if is_property:
            coroutine = function.__get__(*args, **kwargs)  # type: ignore
        else:
            coroutine = function(*args, **kwargs)

        if loop.is_running():
            if threading.current_thread() is threading.main_thread():
                return coroutine
            else:
                if inspect.iscoroutine(coroutine):
                    return asyncio.run_coroutine_threadsafe(coroutine, loop).result()

                if inspect.isasyncgen(coroutine):
                    return asyncio.run_coroutine_threadsafe(consume_generator(coroutine), loop).result()

        return run_sync_ctx(coroutine, loop)

    result = async_to_sync_wrap
    if is_property:
        result = property(result)  # type: ignore
    return result


def async_to_sync(obj: object, name: str, is_property: bool = False) -> None:
    function = getattr(obj, name)

    setattr(obj, name, async_to_sync_wraps(function, is_property=is_property))


def wrap(source: object) -> None:
    for name in dir(source):
        method = getattr(source, name)

        if not name.startswith("_"):
            if inspect.iscoroutinefunction(method) or inspect.isasyncgenfunction(method):
                async_to_sync(source, name)


wrap(RPCProxy)

for source in COINS.values():
    wrap(source)

async_to_sync(COINS["BTC"], "node_id", is_property=True)  # special case: property
async_to_sync(COINS["BTC"], "spec", is_property=True)
async_to_sync(RPCProxy, "spec", is_property=True)
wrap(APIManager)



import functools


class BaseError(Exception):
    """Base error for all errors raised"""


class InvalidEventError(BaseError):
    """Daemon returned unsupported event"""


class NoCurrenciesRegisteredError(BaseError):
    """APIManager has no currencies enabled"""


class LightningDisabledError(BaseError):
    """Lightning is disabled in daemon"""


class ConnectionFailedError(BaseError):
    """Error connecting to the daemon"""


class RequestError(BaseError):
    """Base error for all errors returned from server"""


class UnknownError(RequestError):
    """Unknown error code returned from server"""


@functools.lru_cache()
def generate_exception(exc_name: str) -> type:
    return type(exc_name, (RequestError,), {})


class Errors:
    def __getattr__(self, key: str) -> type:
        return generate_exception(key)


errors = Errors()



from queue import Queue

import pytest

from bitcart.errors import InvalidEventError

test_queue = Queue()


def handler(event, tx):
    test_queue.put(2)


async def async_handler(event, height):
    test_queue.put(3)


@pytest.mark.asyncio
async def test_event_system(btc):
    assert btc.add_event_handler("event", lambda x: test_queue.put(1)) is None

    @btc.on("event")
    def func():
        test_queue.put(2)

    assert btc.add_event_handler(["event1", "event2"], lambda x: test_queue.put(1)) is None
    assert btc.add_event_handler(["new_transaction", "event2"], handler) is None
    assert btc.add_event_handler("new_block", async_handler) is None
    assert set(btc.event_handlers.keys()) == {"event", "event1", "event2", "new_transaction", "new_block"}
    assert btc.event_handlers["new_transaction"] == handler
    assert btc.event_handlers["new_block"] == async_handler
    assert await btc.process_updates({}) is None  # ignoring exceptions
    assert await btc.process_updates([[]]) is None  # ignoring exceptions
    with pytest.raises(InvalidEventError):
        await btc.process_updates([{}])
    with pytest.raises(InvalidEventError):
        await btc.process_updates([{"event": "event"}])

    await btc.process_updates(
        [{"event": "new_transaction"}, {"event": "new_transaction", "tx": "test"}]
    )  # ignoring invalid data, accepting valid
    await btc.process_updates([{"event": "new_block", "height": 1}])
    assert test_queue.qsize() == 2
    assert test_queue.get() == 2
    assert test_queue.get() == 3



import asyncio
import logging
from json import JSONDecodeError
from typing import TYPE_CHECKING, Callable, Dict, Iterable, Optional, Union
from urllib.parse import urljoin

from aiohttp import ClientConnectionError, WSMsgType

from .errors import ConnectionFailedError
from .utils import call_universal

if TYPE_CHECKING:
    from aiohttp import ClientWebSocketResponse

    from .providers.jsonrpcrequests import RPCProxy


class EventDelivery:
    server: "RPCProxy"
    event_handlers: Dict[str, Callable]

    async def process_updates(
        self, updates: Iterable[dict], currency: Optional[str] = None, wallet: Optional[str] = None
    ) -> None:
        raise NotImplementedError()

    async def _register_wallets(self, ws: "ClientWebSocketResponse") -> None:
        raise NotImplementedError()

    async def _start_websocket_processing(
        self, ws: "ClientWebSocketResponse", reconnect_callback: Optional[Callable] = None
    ) -> None:
        await self._register_wallets(ws)
        if reconnect_callback:
            await call_universal(reconnect_callback)
        async for msg in ws:
            if msg.type == WSMsgType.TEXT:
                try:
                    data = msg.json()
                    await self.process_updates(data.get("updates", []), data.get("currency", "BTC"), data.get("wallet"))
                except JSONDecodeError:
                    pass
            elif msg.type == WSMsgType.CLOSED or msg.type == WSMsgType.ERROR:  # pragma: no cover
                break

    async def _start_websocket_inner(self, reconnect_callback: Optional[Callable] = None) -> None:
        async with self.server.session.ws_connect(urljoin(self.server.url, "/ws")) as ws:
            await self._start_websocket_processing(ws, reconnect_callback=reconnect_callback)

    async def _websocket_base_loop(
        self,
        func: Callable,
        reconnect_callback: Optional[Callable] = None,
        force_connect: bool = False,
        auto_reconnect: bool = True,
    ) -> None:
        first = True
        while True:
            try:
                await func(reconnect_callback=reconnect_callback)
            except ClientConnectionError as e:
                if first and not force_connect:
                    raise ConnectionFailedError() from e
            first = False
            if not auto_reconnect:
                break
            await asyncio.sleep(5)  # wait a bit before re-estabilishing a connection # pragma: no cover

    async def start_websocket(
        self, reconnect_callback: Optional[Callable] = None, force_connect: bool = False, auto_reconnect: bool = True
    ) -> None:
        """Start a websocket connection to daemon

        Args:
            reconnect_callback (Optional[Callable], optional): Callback to be called right after
                each succesful connection. Defaults to None.
            force_connect (bool, optional): Whether to try reconnecting even on first failure (handshake)
                to daemon. Defaults to False.
            auto_reconnect (bool, optional): Whether to enable auto-reconnecting on websocket closing. Defaults to True.
        """
        await self._websocket_base_loop(
            self._start_websocket_inner,
            reconnect_callback=reconnect_callback,
            force_connect=force_connect,
            auto_reconnect=auto_reconnect,
        )

    async def poll_updates(self, timeout: Union[int, float] = 1) -> None:  # pragma: no cover
        """Poll updates

        Poll daemon for new transactions in wallet,
        this will block forever in while True loop checking for new transactions

        Example can be found on main page of docs

        Args:
            self (BTC): self
            timeout (Union[int, float], optional): seconds to wait before requesting transactions again. Defaults to 1.

        Raises:
            InvalidEventError: If server sent invalid event name not matching ALLOWED_EVENTS

        Returns:
            None: This function runs forever
        """
        while True:
            try:
                data = await self.server.get_updates()
            except Exception as err:
                logging.error(err)
                await asyncio.sleep(timeout)
                continue
            await self.process_updates(data)
            await asyncio.sleep(timeout)

    def add_event_handler(self, events: Union[Iterable[str], str], func: Callable) -> None:
        """Add event handler to handle event(s) provided

        Args:
            self (BTC): self
            events (Union[Iterable[str], str]): event or events
            func (Callable): function to handle those

        Returns:
            None: None
        """
        if isinstance(events, str):
            events = [events]
        for event in events:
            self.event_handlers[event] = func

    def on(self, events: Union[Iterable[str], str]) -> Callable:
        """Register on event

        Register callback function to be run when event is emmited

        All available events are accessable as:

        >>> btc.ALLOWED_EVENTS
        ['new_block', 'new_transaction']

        Function signature must be

        .. code-block:: python

            def handler(event, **kwargs):

        kwargs sent differ from event to event, as for now
        new_block event sends height kwarg as new block height
        new_transaction event sends tx kwarg as tx_hash of new transaction

        Args:
            self (BTC): self
            events (Union[Iterable[str], str]): event name or list of events for function to be run on

        Returns:
            Callable: It is a decorator
        """

        def wrapper(f: Callable) -> Callable:
            self.add_event_handler(events, f)
            return f

        return wrapper



import inspect
import json
from decimal import Decimal
from typing import Any, Callable

CONVERT_RATE = 100000000


def convert_amount_type(amount: str) -> Decimal:
    """Convert amount from str to Decimal

    Args:
        amount (str): amount

    Returns:
        Decimal
    """
    return Decimal(amount)


def satoshis(amount: Decimal) -> int:
    """Convert amount from bitcoins to satoshis

    Args:
        amount (Decimal): bitcoin amount

    Returns:
        int: same amount in satoshis
    """
    return int(amount * CONVERT_RATE)


def bitcoins(amount: int) -> Decimal:
    """Convert amount from satoshis to bitcoins

    Args:
        amount (int): amount in satoshis

    Returns:
        Decimal: amount in bitcoins
    """
    return Decimal(amount) / Decimal(CONVERT_RATE)


class CustomJSONEncoder(json.JSONEncoder):
    def default(self, obj: Any) -> Any:
        if isinstance(obj, Decimal):
            return str(obj)
        return super().default(obj)  # pragma: no cover


def json_encode(obj: Any) -> Any:
    """json.dumps supporting Decimals

    Args:
        obj (Any): any object

    Returns:
        Any: return value of json.dumps
    """
    return json.dumps(obj, cls=CustomJSONEncoder)


async def call_universal(func: Callable, *args: Any, **kwargs: Any) -> Any:
    """Call a function: async or sync one. All passed arguments are passed to the function too

    Args:
        func (Callable): a function to call: either sync or async one

    Returns:
        Any: function execution result
    """
    result = func(*args, **kwargs)
    if inspect.isawaitable(result):
        await result
    return result



from decimal import Decimal

import pytest

from bitcart import errors

from ...utils import data_check

pytestmark = pytest.mark.asyncio


async def test_balance(btc_wallet):
    attrs = ["confirmed", "unconfirmed", "unmatured", "lightning"]
    balance = await btc_wallet.balance()
    for attr in attrs:
        assert balance[attr] >= 0  # NOTE: unconfirmed can be negative
        assert isinstance(balance[attr], Decimal)


async def test_history(btc_wallet):
    history = await btc_wallet.history()
    data_check(history, "summary", dict)
    data_check(history, "transactions", list, 0)


async def test_payment_request(btc_wallet):
    # request1
    request1_amount = "0.5"
    request1 = await btc_wallet.add_request(request1_amount)
    assert request1[btc_wallet.amount_field] == request1["amount_BTC"] == Decimal(request1_amount)
    # request2
    request2_amount, request2_desc = "0.6", "test description"
    request2 = await btc_wallet.add_request(request2_amount, request2_desc)
    assert request2["amount_BTC"] == Decimal(request2_amount)
    assert request2["message"] == request2_desc
    # get request2
    response2 = await btc_wallet.get_request(request2["address"])
    assert response2["amount_BTC"] == Decimal(request2_amount)
    assert response2["message"] == request2_desc
    # full data structure check
    assert (
        request1.items()
        > {
            "is_lightning": False,
            "amount_BTC": Decimal("0.5"),
            "message": "",
            "expiration": 900,
            "status": 0,
            "status_str": "Expires in 15 minutes",
            "amount_sat": 50000000,
        }.items()
    )
    data_check(request1, "timestamp", int)
    data_check(request1, "address", str)
    data_check(request1, "URI", str)


async def test_insufficient_funds_pay(btc_wallet):
    with pytest.raises(errors.NotEnoughFundsError):
        await btc_wallet.pay_to("1KQah89MJhmhxB9hRQN6DnMSkPH3RUjPev", 0.1)


async def test_fee_and_feerate(btc_wallet):  # can't set both fee and feerate
    with pytest.raises(TypeError):
        await btc_wallet.pay_to("1KQah89MJhmhxB9hRQN6DnMSkPH3RUjPev", 0.1, fee=1, feerate=1)
    with pytest.raises(TypeError):
        await btc_wallet.pay_to_many(
            [("1KQah89MJhmhxB9hRQN6DnMSkPH3RUjPev", 0.1), ("1KQah89MJhmhxB9hRQN6DnMSkPH3RUjPev", 0.1)], fee=1, feerate=1
        )



"""Tests bitcoin methods implementation on mainnet, requires daemon to be running on localhost:5000

If this succeeds, most likely other coins will succeed too
"""
import decimal
import os

import pytest

from bitcart import BTC

from ...utils import data_check

pytestmark = pytest.mark.asyncio


async def test_compare(btc):
    assert btc == BTC()
    assert btc != 1


async def test_help(btc):
    data = await btc.help()
    assert isinstance(data, list)
    assert len(data) > 0
    assert "broadcast" in data
    assert "gettransaction" in data
    assert "help" in data


@pytest.mark.parametrize("currency", ["USD", "RUB", "JPY"])
async def test_rate(btc, currency):
    price = await btc.rate(currency)
    assert price > 0
    assert isinstance(price, decimal.Decimal)


async def test_fiat(btc):
    fiat_currencies = await btc.list_fiat()
    assert ["USD", "BTC", "RUB", "JPY", "EUR"] >= fiat_currencies


@pytest.mark.parametrize("address,expected", [("x", False), ("1KQah89MJhmhxB9hRQN6DnMSkPH3RUjPev", True)])
async def test_electrum_validate_address(btc, address, expected):
    assert await btc.server.validateaddress(address) == expected


@pytest.mark.parametrize(
    "key,expected",
    [
        ("x", False),
        ("gentle expire fatal fashion envelope cheap fury hunt inner copper boy relax", True),
        (
            "xprv9s21ZrQH143K2MFWUoJmzzhhi1SNUqCFeJjbjpWvpvx9T6yxJXUpfW7fPECCNmzHGHQ6qkZYvRGvKu6D2KunXJ2kNccxT6b266F92DbRu35",
            True,
        ),
    ],
)
async def test_validate_key(btc, key, expected):
    assert await btc.validate_key(key) == expected


async def test_get_tx(btc):
    info = await btc.get_tx("0584c650e7b04cd0788832f8340ead4ce654e82127e283c8132a0bcbfabc7a01")
    assert info.items() > {"version": 1, "locktime": 0}.items()
    data_check(info, "confirmations", int)
    assert info["confirmations"] > 0
    data_check(info, "inputs", list, 1)
    assert (
        info["inputs"][0].items()
        >= {
            "prevout_hash": "f5bec7770c30679bd7e91ecc3cf243d43c5a72b28ac8e037eb8c1b5bfe520e27",
            "prevout_n": 0,
            "coinbase": False,
            "scriptSig": "",
            "nsequence": 4294967295,
            "witness": (
                "02473044022030f19cf4b53b0c5d7a9d920240f6daab0cf8a40193f80a8"
                "0d40635dc0ca213af0220487e8714a082f22dba14733f1149bd365767702"
                "f69bc88375d1c47ba8d12138c012103765314d3e5ebb0af504c4993952d9"
                "903648fa3cc645dbaedfd9c40484531836e"
            ),
        }.items()
    )
    data_check(info, "outputs", list, 2)
    assert info["outputs"][0] == {
        "value_sats": 490000,
        "address": "3NyLhw2jKrA8PF2SqgGQu4cigfCg1i6SqH",
        "scriptpubkey": "a914e970f9ca2f7d99b208b455333b0a4ab2b9b7eb3a87",
    }


async def test_config_methods(btc):
    k, v = "x", 1
    await btc.set_config(k, v)
    assert await btc.get_config(k) == v


async def test_get_address(btc):
    txes = await btc.get_address("17ncZMaFQYZYNDycTJc5aydUKaw6oCEUSQ")
    assert isinstance(txes, list)
    tx = txes[0]
    assert tx["tx_hash"] == "74fb8886abb00e66b192172d5fca504337b11af3aac658979355cefe3d51818e"
    assert tx["height"] == 645013
    assert tx["tx"] == await btc.get_tx(tx["tx_hash"])


async def test_create_wallet(btc, tmp_path):
    wallet_path = os.path.join(str(tmp_path), "my_wallet")
    wallet = await btc.server.create(wallet_path=wallet_path)
    assert set(wallet.keys()) == {"seed", "path", "msg"}
    assert wallet["path"] == wallet_path



from setuptools import find_packages, setup


def main() -> None:
    setup(
        name="bitcart",
        packages=find_packages(),
        version="1.3.2.0",
        license="LGPLv3+",
        description="BitcartCC coins support library",
        long_description=open("README.md").read(),
        long_description_content_type="text/markdown",
        author="MrNaif2018",
        author_email="chuff184@gmail.com",
        url="https://github.com/bitcartcc/bitcart-sdk",
        keywords=["electrum", "daemon", "bitcart", "bitcartcc"],
        install_requires=["jsonrpcclient[aiohttp]", "aiohttp<4.0.0"],
        extras_require={"proxy": ["aiohttp_socks"]},
        classifiers=[
            "Development Status :: 4 - Beta",
            "Intended Audience :: Developers",
            "Topic :: Software Development :: Build Tools",
            "License :: OSI Approved :: GNU Lesser General Public License v3 or later (LGPLv3+)",
            "Programming Language :: Python :: 3",
            "Programming Language :: Python :: 3.7",
            "Programming Language :: Python :: 3.8",
            "Programming Language :: Python :: 3.9",
        ],
    )


if __name__ == "__main__":
    main()



import asyncio

import pytest

from bitcart import errors

from .utils import data_check, run_shell

pytestmark = pytest.mark.asyncio

BTC_ADDRESS = (
    "bcrt1qe0ppfnuz6wjw3vn8jefn8p4fxmyn7tqxkjt557"  # can be got by run_shell(["newaddress"]) or regtest_wallet.add_request()
)

TEST_PARAMS = [
    (None, None, True),
    (None, None, False),
    (0.01, None, True),
    (0.01, None, False),
    (lambda size, default_fee: default_fee, None, True),
    (lambda size, default_fee: default_fee, None, False),
    (lambda size, default_fee: default_fee / 0, None, True),  # silent fallback to default fee if exception is raised
    (lambda size, default_fee: default_fee / 0, None, False),
    (None, 5, True),
    (None, 5, False),
]


def setup_module(module):
    run_shell(["startup", "5", BTC_ADDRESS])  # make blockchain mature and add funds


@pytest.fixture
async def wait_for_balance(regtest_wallet):
    while True:
        balance = await regtest_wallet.balance()
        balance = balance["confirmed"] + balance["unconfirmed"]
        if balance < 5:
            await asyncio.sleep(1)
        else:
            break


@pytest.fixture
async def wait_for_utxos(regtest_wallet):
    while True:
        utxos = [utxo for utxo in await regtest_wallet.server.listunspent() if utxo["height"] != -2]
        # ensure we have nonlocal utxos to use to open channel
        if len(utxos) == 0:
            await asyncio.sleep(1)
        else:
            break


def check_tx(tx, broadcast):
    assert isinstance(tx, str)
    if broadcast:
        assert len(tx) == 64
    # if it is not broadcast, it returns raw transaction


@pytest.mark.parametrize("fee,feerate,broadcast", TEST_PARAMS)
async def test_payment_to_single(regtest_wallet, fee, feerate, broadcast, wait_for_balance):
    check_tx(
        await regtest_wallet.pay_to(
            "bcrt1qx4y7d5wt9cn585cqyxc7899khtz2dsl0qnufkz",
            0.1,
            fee=fee,
            feerate=feerate,
            broadcast=broadcast,
        ),
        broadcast,
    )


@pytest.mark.parametrize("fee,feerate,broadcast", TEST_PARAMS)
async def test_payment_to_many(regtest_wallet, fee, feerate, broadcast, wait_for_balance):
    check_tx(
        await regtest_wallet.pay_to_many(
            [("bcrt1qx4y7d5wt9cn585cqyxc7899khtz2dsl0qnufkz", 0.1), ("bcrt1qx4y7d5wt9cn585cqyxc7899khtz2dsl0qnufkz", 0.1)],
            fee=fee,
            feerate=feerate,
            broadcast=broadcast,
        ),
        broadcast,
    )
    check_tx(
        await regtest_wallet.pay_to_many(
            [
                {"address": "bcrt1qx4y7d5wt9cn585cqyxc7899khtz2dsl0qnufkz", "amount": 0.1},
                {"address": "bcrt1qx4y7d5wt9cn585cqyxc7899khtz2dsl0qnufkz", "amount": 0.1},
            ],
            fee=fee,
            feerate=feerate,
            broadcast=broadcast,
        ),
        broadcast,
    )


async def test_open_channel(regtest_wallet, regtest_node_id, wait_for_utxos):
    # works only with nonlocal balances
    result = await regtest_wallet.open_channel(regtest_node_id, 0.002)
    assert isinstance(result, str)
    assert len(result) == 66
    assert ":" in result
    splitted = result.split(":")
    assert len(splitted) == 2
    int(splitted[1])  # integer
    run_shell(["newblocks", "3"])


async def test_list_channels(regtest_wallet, regtest_node_id):
    pubkey = regtest_node_id.split("@")[0]
    result = await regtest_wallet.list_channels()
    assert isinstance(result, list)
    assert len(result) > 0
    channel = result[-1]  # last channel is last in the list
    assert "short_channel_id" in channel
    assert isinstance(channel["short_channel_id"], str) or channel["short_channel_id"] is None
    data_check(channel, "channel_id", str, 64)
    data_check(channel, "channel_point", str, 66)
    data_check(channel, "peer_state", str)
    assert channel["peer_state"] == "GOOD"
    data_check(channel, "state", str)
    assert channel["remote_pubkey"] == pubkey
    assert channel["local_balance"] == 200000
    assert channel["remote_balance"] == 0
    data_check(channel, "local_reserve", int)
    data_check(channel, "remote_reserve", int)
    data_check(channel, "local_unsettled_sent", int)
    data_check(channel, "remote_unsettled_sent", int)
    assert channel["local_reserve"] == channel["remote_reserve"] == 2000
    assert channel["local_unsettled_sent"] == channel["remote_unsettled_sent"] == 0


async def test_lnpay(regtest_wallet):
    with pytest.raises(errors.InvalidLightningInvoiceError):
        assert not await regtest_wallet.lnpay("")
    response = await regtest_wallet.lnpay((await regtest_wallet.add_invoice(0.1))["invoice"])
    assert isinstance(response, dict)
    assert (
        response.items()
        > {
            "success": False,
            "preimage": None,
            "log": [["None", "N/A", "No path found"]],
        }.items()
    )
    data_check(response, "payment_hash", str)


async def test_close_channel(regtest_wallet):
    channels = await regtest_wallet.list_channels()
    channel_id = channels[-1]["channel_point"]  # last channel is last in the list
    assert isinstance(await regtest_wallet.close_channel(channel_id), str)



from typing import Any

from .btc import BTC


class BCH(BTC):
    coin_name = "BCH"
    friendly_name = "Bitcoin Cash"
    RPC_URL = "http://localhost:5004"
    AMOUNT_FIELD = "amount (BCH)"

    async def history(self) -> dict:  # pragma: no cover
        return await self.server.history()  # type: ignore

    async def _add_request(self, *args: Any, **kwargs: Any) -> dict:  # pragma: no cover
        return await self.server.addrequest(*args, **kwargs)  # type: ignore
